# classifier plugin for generating hierarchical classifications

package Hierarchy;

use util;
use cfgread;

sub new {
    my ($class, @options) = @_;

    if (!defined @options || (scalar @options < 2)) {
	die "Error in options passed to Hierarchy classification\n" .
	    "Usage: Hierarchy subjectfile metadata\n";
    }

    my $subjectfile = &util::filename_cat($ENV{'GSDLCOLLECTDIR'},"etc", $options[0]);
    if (!-e $subjectfile) {
	my $collfile = $subjectfile;
	$subjectfile = &util::filename_cat($ENV{'GSDLHOME'},"etc", $options[0]);
	if (!-e $subjectfile) {
	    die "Error: Can't locate subject file $options[0]\n" .
		"This file should be in $collfile or $subjectfile\n";
	}
    }

    return bless {
	'OID'=>"NULL",
	'descriptorlist'=>{},
	'locatorlist'=>{},
	'subjectfile' => $subjectfile,
	'metaname' => $options[1]
	}, $class;
}

sub init {
    my $self = shift (@_);
    my $contenthash = {};

    # read in the subject file
    my $list = &cfgread::read_cfg_file ($self->{'subjectfile'}, undef, '^[^#]?\w');
    
    foreach $descriptor (keys (%$list)) {
	$self->{'descriptorlist'}->{$descriptor} = $list->{$descriptor}->[0];
	unless (defined $self->{'locatorlist'}->{$list->{$descriptor}->[0]}) {
	    $self->{'locatorlist'}->{$list->{$descriptor}->[0]}->{'title'} = $list->{$descriptor}->[1];
	    $self->{'locatorlist'}->{$list->{$descriptor}->[0]}->{'contents'} = [];
	}
	
	my $OID = $list->{$descriptor}->[0];
	if ($OID =~ s/\.(\d+)$//) {
	    # $OID is now the parents OID
	    my $child = $1;
	    $contenthash->{$OID} = [] unless defined $contenthash->{$OID};
	    push (@{$contenthash->{$OID}}, $child);
	}
    }

    # need to shove each OID into the contents of its parent
    # do this separately in case classifications aren't in order in 
    # subject file
    foreach $OID (keys (%$contenthash)) {
	push (@{$self->{'locatorlist'}->{$OID}->{'contents'}}, 
	      sort numerically @{$contenthash->{$OID}});
	map { $_ = "\".$_"; } @{$self->{'locatorlist'}->{$OID}->{'contents'}};
    }
}

sub set_OID {
    my $self = shift (@_);
    my ($OID) = @_;

    $self->{'OID'} = $OID;
}

sub get_OID {
    my $self = shift (@_);
    return $self->{'OID'};
}

sub classify {
    my $self = shift (@_);
    my ($doc_obj) = @_;

    my $doc_OID = $doc_obj->get_OID();
    my $metadata = $doc_obj->get_metadata ($doc_obj->get_top_section(),
					   $self->{'metaname'});

    foreach $metaelement (@$metadata) {
	if ((defined $self->{'descriptorlist'}->{$metaelement}) &&
	    (defined $self->{'locatorlist'}->{$self->{'descriptorlist'}->{$metaelement}})) {
	    push (@{$self->{'locatorlist'}->{$self->{'descriptorlist'}->{$metaelement}}->{'contents'}}, $doc_OID);
	}
    }
}

sub output_classify_info {
    my $self = shift (@_);
    my ($handle) = @_;
    my $topcontents = [];

    my $list = $self->{'locatorlist'};

    foreach $OID (keys (%$list)) {
	$self->output_entry ($handle, "$self->{'OID'}.$OID", 
			     $list->{$OID}->{'title'}, $list->{$OID}->{'contents'});
	push (@$topcontents, $OID) if $OID !~ /\./;
    }

    # top level of classification
    map { $_ = "\".$_"; } sort numerically @$topcontents;
    $self->output_entry ($handle, $self->{'OID'}, $self->{'metaname'}, 
			 $topcontents, "Hierarchy");
}

sub output_entry {
    my $self = shift (@_);
    my ($handle, $OID, $title, $contentsref, $classifytype) = @_;
    
    print $handle "[$OID]\n";
    print $handle "<doctype>classify\n";
    print $handle "<hastxt>0\n";
    print $handle "<Title>$title\n";
    print $handle "<classifytype>$classifytype\n" if defined $classifytype;
    if (scalar @$contentsref) {
	print $handle "<contains>", join (";", @$contentsref), "\n";
    }
    print $handle '-' x 70, "\n";
}

sub numerically { $a <=> $b; }

1;
