# MGBuilder object
#

package mgbuilder;

use classify;
use cfgread;
use colcfg;
use plugin;
use util;

$maxdocsize = 12000;

%wanted_index_files = ('td'=>1,
		       't'=>1,
		       'idb'=>1,
		       'ib1'=>1,
		       'ib2'=>1,
		       'ib3'=>1,
		       'i'=>1,
		       'ip'=>1,
		       'tiw'=>1,
		       'wa'=>1);


sub new {
    my ($class, $collection, $source_dir, $build_dir, $verbosity, $newgdbm) = @_;

    # create an mgbuilder object
    my $self = bless {'collection'=>$collection,
		      'source_dir'=>$source_dir,
		      'build_dir'=>$build_dir,
		      'verbosity'=>$verbosity,
		      'newgdbm'=>$newgdbm}, $class;


    # read in the collection configuration file
    my $colcfgname = "$ENV{'GSDLCOLLECTDIR'}/etc/collect.cfg";
    if (!-e $colcfgname) {
	die "mgbuilder::new - couldn't find collect.cfg for collection $collection\n";
    }
    $self->{'collect_cfg'} = &colcfg::read_collect_cfg ($colcfgname);

    # sort out subcollection indexes
    if (defined $self->{'collect_cfg'}->{'indexsubcollections'}) {
	my $indexes = $self->{'collect_cfg'}->{'indexes'};
	$self->{'collect_cfg'}->{'indexes'} = [];
	foreach $subcollection (@{$self->{'collect_cfg'}->{'indexsubcollections'}}) {
	    foreach $index (@$indexes) {
		push (@{$self->{'collect_cfg'}->{'indexes'}}, "$index:$subcollection");
	    }
	}
    }

    # sort out language subindexes
    if (defined $self->{'collect_cfg'}->{'languages'}) {
	my $indexes = $self->{'collect_cfg'}->{'indexes'};
	$self->{'collect_cfg'}->{'indexes'} = [];
	foreach $language (@{$self->{'collect_cfg'}->{'languages'}}) {
	    foreach $index (@$indexes) {
		push (@{$self->{'collect_cfg'}->{'indexes'}}, "$index:$language");
	    }
	}
    }

    # get the list of plugins for this collection
    my @plugins = (); # some good choice of plugins .... ????
    if (defined $self->{'collect_cfg'}->{'plugins'}) {
	@plugins = @{$self->{'collect_cfg'}->{'plugins'}};
    }
    

    # load all the plugins
    $self->{'pluginfo'} = &plugin::load_plugins ($collection, \@plugins);
    if (scalar(@{$self->{'pluginfo'}}) == 0) {
	print STDERR "No plugins were loaded.\n";
	die "\n";
    }

    # load all the classifiers
    if ($self->{'newgdbm'}) {
	$self->{'classifiers'} = [];
	if (open (COLCFG, $colcfgname)) {
	    while (defined ($line = &cfgread::read_cfg_line('mgbuilder::COLCFG'))) {
		if (scalar(@$line) >= 2) {
		    my $key = shift (@$line);
		    if ($key eq "classify") {
			my $classinfo = &classify::load_classifier($line);
			push (@{$self->{'classifiers'}}, $classinfo) 
			    if defined $classinfo;
		    }
		}
	    }
	    close (COLCFG);
	} else {
	    print STDERR "mgbuilder::new couldn't read the cfg file $colcfgname\n";
	    print STDERR "               no classifiers were loaded\n";
	}
    }

    # set the classifytype to use for displaying documents - if the doctype field hasn't 
    # been set in the collect.cfg then the receptionist currently defaults to displaying
    # documents as 'Book'
    if ($self->{'newgdbm'}) {
	if (open (COLCFG, $colcfgname)) {
	    while (defined ($line = &cfgread::read_cfg_line('mgbuilder::COLCFG'))) {
		if (scalar(@$line) == 2) {
		    my $key = shift (@$line);
		    if ($key eq "doctype") {
			$self->{'classifytype'} = shift (@$line);
		    }
		}
	    }
	    close (COLCFG);
	}
    }

    # load up the document processor for building
    # if a buildproc class has been created for this collection, use it
    # otherwise, use the mg buildproc
    my ($buildprocdir, $buildproctype);
    if (-e "$ENV{'GSDLCOLLECTDIR'}/perllib/${collection}buildproc.pm") {
	$buildprocdir = "$ENV{'GSDLCOLLECTDIR'}/perllib";
	$buildproctype = "${collection}buildproc";
    } else {
	$buildprocdir = "$ENV{'GSDLHOME'}/perllib";
	$buildproctype = "mgbuildproc";
    }
    require "$buildprocdir/$buildproctype.pm";

    eval("\$self->{'buildproc'} = new $buildproctype(\$collection, " .
	 "\$source_dir, \$build_dir, \$verbosity, \$newgdbm)");
    die "$@" if $@;


    return $self;
}

sub init {
    my $self = shift (@_);

    # remove any old builds
    &util::rm_r($self->{'build_dir'});
    &util::mk_all_dir($self->{'build_dir'});
        
    # make the text directory
    my $textdir = "$self->{'build_dir'}/text";
    &util::mk_all_dir($textdir);
}

sub compress_text {
    my $self = shift (@_);
    my ($textindex) = @_;
    my $exedir = "$ENV{'GSDLHOME'}/bin/$ENV{'GSDLOS'}";
    my $exe = &util::get_os_exe ();

    &util::mk_all_dir (&util::filename_cat($self->{'build_dir'}, "text"));
    my $basefilename = "text/$self->{'collection'}";
    my $fulltextprefix = &util::filename_cat ($self->{'build_dir'}, $basefilename);

    my $osextra = "";
    if ($ENV{'GSDLOS'} =~ /^windows$/i) {
	$fulltextprefix =~ s/\//\\/g;
    } else {
	$osextra = " -d /";
    }

    print STDERR "\n*** creating the compressed text\n" if ($self->{'verbosity'} >= 1);

    # set up the document processor
    $self->{'buildproc'}->set_output_handle ('mgbuilder::PIPEOUT');
    $self->{'buildproc'}->set_mode ('text');
    $self->{'buildproc'}->set_index ($textindex);
    $self->{'buildproc'}->set_indexing_text (0);
    
    # collect the statistics for the text
    # -b $maxdocsize sets the maximum document size to be 12 meg
    print STDERR "\n    collecting text statistics\n"  if ($self->{'verbosity'} >= 1);
    if (!-e "$exedir/mg_passes$exe" || !open (PIPEOUT, 
            "| $exedir/mg_passes$exe -f $fulltextprefix -b $maxdocsize -T1 $osextra")) {
	die "mgbuilder::compress_text - couldn't run $exedir/mg_passes$exe\n";
    }
    $self->{'buildproc'}->reset();
    &plugin::read ($self->{'pluginfo'}, $self->{'source_dir'}, 
		   "", {}, $self->{'buildproc'});
    close (PIPEOUT);

    # create the compression dictionary
    # the compression dictionary is built by assuming the stats are from a seed
    # dictionary (-S), if a novel word is encountered it is spelled out (-H),
    # and the resulting dictionary must be less than 5 meg with the most frequent
    # words being put into the dictionary first (-2 -k 5120)
    print STDERR "\n    creating the compression dictionary\n"  if ($self->{'verbosity'} >= 1);
    if (!-e "$exedir/mg_compression_dict$exe") {
	die "mgbuilder::compress_text - couldn't run $exedir/mg_compression_dict$exe\n";
    }
    system ("$exedir/mg_compression_dict$exe -f $fulltextprefix -S -H -2 -k 5120 $osextra");

    # compress the text
    # -b $maxdocsize sets the maximum document size to be 12 meg
    print STDERR "\n    compressing the text\n"  if ($self->{'verbosity'} >= 1);
    if (!-e "$exedir/mg_passes$exe" || !open (PIPEOUT, 
            "| $exedir/mg_passes$exe -f $fulltextprefix -b $maxdocsize -T2 $osextra")) {
	die "mgbuilder::compress_text - couldn't run $exedir/mg_passes$exe\n";
    }
    $self->{'buildproc'}->reset();
    &plugin::read ($self->{'pluginfo'}, $self->{'source_dir'}, 
		   "", {}, $self->{'buildproc'});
    close (PIPEOUT);
}

sub build_indexes {
    my $self = shift (@_);
    my $indexes = $self->{'collect_cfg'}->{'indexes'};

    # create the mapping between the index descriptions 
    # and their directory names
    $self->{'index_mapping'} = $self->create_index_mapping ($indexes);

    # build each of the indexes
    foreach $index (@$indexes) {
	print STDERR "\n*** building index $index in subdirectory " . 
	    "$self->{'index_mapping'}->{$index}\n" if ($self->{'verbosity'} >= 1);
	$self->build_index($index);
    }
}

# creates directory names for each of the index descriptions
sub create_index_mapping {
    my $self = shift (@_);
    my ($indexes) = @_;

    my %mapping = ();
    $mapping{'indexmaporder'} = [];
    $mapping{'subcollectionmaporder'} = [];
    $mapping{'languagemaporder'} = [];
    
    # dirnames is used to check for collisions. Start this off
    # with the manditory directory names
    my %dirnames = ('text'=>'text',
		    'extra'=>'extra');
    my %pnames = ('index' => '', 'subcollection' => '', 'languages' => '');

    foreach $index (@$indexes) {
	my ($level, $gran, $subcollection, $languages) = split (":", $index);

	# the directory name starts with the first character of the index level
	my ($pindex) = $level =~ /^(.)/;

	# next comes a processed version of the index
	$pindex .= $self->process_field ($gran); 
	$pindex = lc ($pindex);

	# next comes a processed version of the subcollection if there is one.
	my $psub = $self->process_field ($subcollection);
	$psub = lc ($psub);

	# next comes a processed version of the language if there is one.
	my $plang = $self->process_field ($languages);
	$plang = lc ($plang);

	my $dirname = $pindex . $psub . $plang;

	# check to be sure all index names are unique
	while (defined ($dirnames{$dirname})) {
	    $dirname = $self->make_unique (\%pnames, $index, \$pindex, \$psub, \$plang);
	}

	# store the mapping orders as well as the maps
	if (!defined $mapping{'indexmap'}{"$level:$gran"}) {
	    $mapping{'indexmap'}{"$level:$gran"} = $pindex;
	    push (@{$mapping{'indexmaporder'}}, "$level:$gran");
	}
	if ($psub =~ /\w/ && !defined ($mapping{'subcollectionmap'}{$subcollection})) {
	    $mapping{'subcollectionmap'}{$subcollection} = $psub;
	    push (@{$mapping{'subcollectionmaporder'}}, $subcollection);
	}
	if ($plang =~ /\w/ && !defined ($mapping{'languagemap'}{$languages})) {
	    $mapping{'languagemap'}{$languages} = $plang;
	    push (@{$mapping{'languagemaporder'}}, $language);
	}
	$mapping{$index} = $dirname;
	$dirnames{$dirname} = $index;
	$pnames{'index'}{$pindex} = "$level:$gran";
	$pnames{'subcollection'}{$psub} = $subcollection;
	$pnames{'languages'}{$plang} = $languages;
    }

    return \%mapping;
}

# returns a processed version of a field.
# if the field has only one component the processed
# version will contain the first character and next consonant
# of that componant - otherwise it will contain the first 
# character of the first two components 
sub process_field {
    my $self = shift (@_);
    my ($field) = @_;
 
    return "" unless (defined ($field) && $field =~ /\w/);

    my @components = split /,/, $field;
    if (scalar @components >= 2) {
	splice (@components, 2);
	map {s/^(.).*$/$1/;} @components;
	return join("", @components);
    } else {
	my ($a, $b) = $field =~ /^(.).*?([bcdfghjklmnpqrstvwxyz])/i;
	($a, $b) = $field =~ /^(.)(.)/ unless defined $a && defined $b;
	return "$a$b";
    }
}

sub make_unique {
    my $self = shift (@_);
    my ($namehash, $index, $indexref, $subref, $langref) = @_;
    my ($level, $gran, $subcollection, $languages) = split (":", $index);

    if ($namehash->{'index'}->{$$indexref} ne "$level:$gran") {
	$self->get_next_version ($indexref);
    } elsif ($namehash->{'subcollection'}->{$$subref} ne $subcollection) {
	$self->get_next_version ($subref);
    } elsif ($namehash->{'languages'}->{$$langref} ne $languages) {
	$self->get_next_version ($langref);
    }
    return "$$indexref$$subref$$langref";
}	

sub get_next_version {
    my $self = shift (@_);
    my ($nameref) = @_;

    if ($$nameref =~ /(\d\d)$/) {
	my $num = $1; $num ++;
	$$nameref =~ s/\d\d$/$num/;
    } elsif ($$nameref =~ /(\d)$/) {
	my $num = $1;
	if ($num == 9) {$$nameref =~ s/\d\d$/10/;}
	else {$num ++; $$nameref =~ s/\d$/$num/;}
    } else {
	$$nameref =~ s/.$/0/;
    }
}

sub build_index {
    my $self = shift (@_);
    my ($index) = @_;

    # get the full index directory path and make sure it exists
    my $indexdir = $self->{'index_mapping'}->{$index};
    &util::mk_all_dir (&util::filename_cat($self->{'build_dir'}, $indexdir));
    my $fullindexprefix = &util::filename_cat ($self->{'build_dir'}, $indexdir, 
					       $self->{'collection'});
    my $fulltextprefix = &util::filename_cat ($self->{'build_dir'}, "text", 
					       $self->{'collection'});

    # get any os specific stuff
    my $exedir = "$ENV{'GSDLHOME'}/bin/$ENV{'GSDLOS'}";
    my $exe = &util::get_os_exe ();
    my $osextra = "";
    if ($ENV{'GSDLOS'} =~ /^windows$/i) {
	$fullindexprefix =~ s/\//\\/g;
    } else {
	$osextra = " -d /";
    }

    # get the index level from the index description
    # the index will be level 2 unless we are building a
    # paragraph level index
    my $index_level = 2;
    $index_level = 3 if $index =~ /^paragraph/i;

    # get the index expression if this index belongs
    # to a subcollection
    my $indexexparr = [];
    my ($level, $fields, $subcollection) = split (":", $index);
    my @subcollections = ();
    @subcollections = split /,/, $subcollection if (defined $subcollection);

    foreach $subcollection (@subcollections) {
	if (defined ($self->{'collect_cfg'}->{'subcollection'}->{$subcollection})) {
	    push (@$indexexparr, $self->{'collect_cfg'}->{'subcollection'}->{$subcollection});
	} 
    }
    
    # add expressions for languages if this index belongs to
    # a language subcollection
    foreach $language (@{$self->{'collect_cfg'}->{'languages'}}) {
	if ($language =~ s/^\!//) {
	    push (@$indexexparr, "!Language/$language/");
	} else {
	    push (@$indexexparr, "Language/$language/");
	}
    }
    
    # set up the document processor
    $self->{'buildproc'}->set_output_handle ('mgbuilder::PIPEOUT');
    $self->{'buildproc'}->set_mode ('text');
    $self->{'buildproc'}->set_index ($index, $indexexparr);
    $self->{'buildproc'}->set_indexing_text (1);


    # Build index dictionary. Uses verbatim stem method
    print STDERR "\n    creating index dictionary\n"  if ($self->{'verbosity'} >= 1);
    if (!-e "$exedir/mg_passes$exe" || !open (PIPEOUT, 
	    "| $exedir/mg_passes$exe -f $fullindexprefix -b $maxdocsize " .
	    "-$index_level -m 32 -s 0 -G -t 10 -N1 $osextra")) {
	die "mgbuilder::build_index - couldn't run $exedir/mg_passes$exe\n";
    }
    $self->{'buildproc'}->reset();
    &plugin::read ($self->{'pluginfo'}, $self->{'source_dir'}, 
		   "", {}, $self->{'buildproc'});
    close (PIPEOUT);

    # create the perfect hash function
    if (!-e "$exedir/mg_perf_hash_build$exe") {
	die "mgbuilder::build_index - couldn't run $exedir/mg_perf_hash_build$exe\n";
    }
    system ("$exedir/mg_perf_hash_build$exe -f $fullindexprefix $osextra");

    # invert the text
    print STDERR "\n    inverting the text\n"  if ($self->{'verbosity'} >= 1);
    if (!-e "$exedir/mg_passes$exe" || !open (PIPEOUT, 
	    "| $exedir/mg_passes$exe -f $fullindexprefix -b $maxdocsize " .
	    "-$index_level -c 3 -G -t 10 -N2 $osextra")) {
	die "mgbuilder::build_index - couldn't run $exedir/mg_passes$exe\n";
    }
    $self->{'buildproc'}->reset();
    &plugin::read ($self->{'pluginfo'}, $self->{'source_dir'}, 
		   "", {}, $self->{'buildproc'});
    close (PIPEOUT);

    # create the weights file
    print STDERR "\n    create the weights file\n"  if ($self->{'verbosity'} >= 1);
    if (!-e "$exedir/mg_weights_build$exe") {
	die "mgbuilder::build_index - couldn't run $exedir/mg_weights_build$exe\n";
    }
    system ("$exedir/mg_weights_build$exe -f $fullindexprefix -t $fulltextprefix $osextra");

    # create 'on-disk' stemmed dictionary
    print STDERR "\n    creating 'on-disk' stemmed dictionary\n"  if ($self->{'verbosity'} >= 1);
    if (!-e "$exedir/mg_invf_dict$exe") {
	die "mgbuilder::build_index - couldn't run $exedir/mg_invf_dict$exe\n";
    }
    system ("$exedir/mg_invf_dict$exe -f $fullindexprefix $osextra");


    # creates stem index files for the various stemming methods
    print STDERR "\n    creating stem indexes\n"  if ($self->{'verbosity'} >= 1);
    if (!-e "$exedir/mg_stem_idx$exe") {
	die "mgbuilder::build_index - couldn't run $exedir/mg_stem_idx$exe\n";
    }
    system ("$exedir/mg_stem_idx$exe -b 4096 -s1 -f $fullindexprefix $osextra");
    system ("$exedir/mg_stem_idx$exe -b 4096 -s2 -f $fullindexprefix $osextra");
    system ("$exedir/mg_stem_idx$exe -b 4096 -s3 -f $fullindexprefix $osextra");

    
    # remove unwanted files
    my $tmpdir = &util::filename_cat ($self->{'build_dir'}, $indexdir);
    opendir (DIR, $tmpdir) || die
	"mgbuilder::build_index - couldn't read directory $tmpdir\n";
    foreach $file (readdir(DIR)) {
	next if $file =~ /^\./;
	my ($suffix) = $file =~ /\.([^\.]+)$/;
	if (defined $suffix && !defined $wanted_index_files{$suffix}) {
	    # delete it!
#	    print STDERR "deleting $file\n";
	    &util::rm (&util::filename_cat ($tmpdir, $file));
	}
    }
    closedir (DIR);
}

sub make_infodatabase {
    my $self = shift (@_);
    my $textdir = &util::filename_cat($self->{'build_dir'}, "text");
    &util::mk_all_dir ($textdir);

    # get db name
    my $dbext = ".bdb";
    $dbext = ".ldb" if &util::is_little_endian();
    my $fulldbname = &util::filename_cat ($textdir, "$self->{'collection'}$dbext");
    $fulldbname =~ s/\//\\/g if ($ENV{'GSDLOS'} =~ /^windows$/i);

    my $exedir = "$ENV{'GSDLHOME'}/bin/$ENV{'GSDLOS'}";
    my $exe = &util::get_os_exe ();

    print STDERR "\n*** creating the info database\n" if ($self->{'verbosity'} >= 1);

    # init all the classifiers
    if ($self->{'newgdbm'}) {
	&classify::init_classifiers ($self->{'classifiers'});
    }

    # set up the document processor
    $self->{'buildproc'}->set_output_handle ('mgbuilder::PIPEOUT');
    if ($self->{'newgdbm'}) {
	$self->{'buildproc'}->set_mode ('newinfodb');
	$self->{'buildproc'}->set_classifiers ($self->{'classifiers'});
    } else {
	$self->{'buildproc'}->set_mode ('infodb');
    }
    $self->{'buildproc'}->set_indexing_text (0);
    
    # create the infodatabase
    if (!-e "$exedir/txt2db$exe" || !open (PIPEOUT, 
            "| $exedir/txt2db$exe $fulldbname")) {
	die "mgbuilder::make_infodatabase - couldn't run $exedir/txt2db$exe\n";
    }
    $self->{'buildproc'}->reset();

    my $metadata = {};
    if (defined $self->{'classifytype'} && $self->{'classifytype'} =~ /\w/) {
	$metadata->{'classifytype'} = $self->{'classifytype'};
    }
    &plugin::read ($self->{'pluginfo'}, $self->{'source_dir'}, 
		   "", $metadata, $self->{'buildproc'});

    # output classification information
    if ($self->{'newgdbm'}) {
	&classify::output_classify_info ($self->{'classifiers'}, 'mgbuilder::PIPEOUT');
    }

    close (PIPEOUT);
}

sub make_auxiliary_files {
    my $self = shift (@_);
    my ($index);
    my %build_cfg = ();

    print STDERR "\n*** creating auxiliary files \n" if ($self->{'verbosity'} >= 1);

    # get the text directory
    &util::mk_all_dir ($self->{'build_dir'});

    # store the build date
    $build_cfg->{'builddate'} = time;

    # store the number of documents and number of bytes
    $build_cfg->{'numdocs'} = $self->{'buildproc'}->get_num_docs();
    $build_cfg->{'numbytes'} = $self->{'buildproc'}->get_num_bytes();

    # store the mapping between the index names and the directory names
    my @indexmap = ();
    foreach $index (@{$self->{'index_mapping'}->{'indexmaporder'}}) {
	push (@indexmap, "$index\-\>$self->{'index_mapping'}->{'indexmap'}->{$index}");
    }
    $build_cfg->{'indexmap'} = \@indexmap;

    my @subcollectionmap = ();
    foreach $subcollection (@{$self->{'index_mapping'}->{'subcollectionmaporder'}}) {
	push (@subcollectionmap, "$subcollection\-\>$self->{'index_mapping'}->".
	      "{'subcollectionmap'}->{$subcollection}");
    }
    $build_cfg->{'subcollectionmap'} = \@subcollectionmap if scalar (@subcollectionmap);

    my @languagemap = ();
    foreach $language (@{$self->{'index_mapping'}->{'languagemaporder'}}) {
	push (@languagemap, "$language\-\>$self->{'index_mapping'}->{'languagemap'}->".
	      "{$language}");
    }
    $build_cfg->{'languagemap'} = \@languagemap if scalar (@languagemap);

    # write out the build information
    &cfgread::write_cfg_file("$self->{'build_dir'}/build.cfg", $build_cfg,
			     '^(builddate|numdocs|numbytes)$', '^(indexmap|subcollectionmap|languagemap)$');

}

sub deinit {
    my $self = shift (@_);
}


1;


